/*
    remote-sensor, a wireless sensor data collection system
    Copyright (C) 2008  christian mock <cm@tahina.priv.at>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#define F_CPU 12000000UL

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "usbdrv.h"
#include "rf12.h"

// min and max size of received packets
#define PKT_LEN_MIN 2
#define PKT_LEN_MAX 16

// number of receive buffers, must be power of 2
#define NBUFFERS 4
#define BUFFERMASK (NBUFFERS - 1)

// indicator LEDs:
#define LED_R (_BV(PC2))
#define LED_Y (_BV(PC3))
#define LED_G (_BV(PC4))
#define RED_OFF() (PORTC |= LED_R)
#define YELLOW_OFF() (PORTC |= LED_Y)
#define GREEN_OFF() (PORTC |= LED_G)
#define RED_ON() (PORTC &= ~LED_R)
#define YELLOW_ON() (PORTC &= ~LED_Y)
#define GREEN_ON() (PORTC &= ~LED_G)


static uint8_t replyBuffer[PKT_LEN_MAX];

// state machine for reception of wireless packets
// (we can't just do a blocking read, since usbPoll() must be called
// regularily)
// states:
// STATE_INIT: initialize the SM
// STATE_IDLE: we're waiting for the beginning of a packet
// STATE_PKT_READING: we're reading len bytes
// STATE_PKT_READING_DONE: we've read all data bytes, wait for the final cksum
#define STATE_INIT 0
#define STATE_IDLE 1
#define STATE_READING 2
#define STATE_READING_DONE 3

// one iteration of the main loop takes about 35s,
// one packet of 20bytes @9600 takes about 19ms, so...
#define RCV_TIMEOUT 600

struct {
  uint8_t state;		// state machine state
  uint8_t len;			// length byte received
  uint8_t cksum;		// checksum accumulator
  uint8_t i;			// current byte
  uint16_t timeout;		// timeout down-counter
  uint8_t *buf;			// pointer to packet buffer (buffers, below)
} pkt_state;

struct {
  uint16_t ok;			// packets received OK
  uint16_t e_cksum;		// checksum errors
  uint16_t e_to_reading;	// timeouts in READING
  uint16_t e_to_reading_done;	// timeouts in READING_DONE
  uint16_t e_bad_len;		// len out of range
  uint16_t e_queue_full;	// USB queue full
} stats;

// receive buffers and the lenghts of the packets therein
uint8_t buffers[NBUFFERS][PKT_LEN_MAX];
uint8_t buffers_len[NBUFFERS];
uint8_t rbuf, ubuf;	// radio (write) and USB (read) buffer pointer

uint16_t usbStatusLED; // counter to flash the red LED on USB activity
uint16_t rxStatusLED;		// same for radio rcv

void receive_packet() {
  uint8_t rcv_cksum;

  switch (pkt_state.state) {
  case STATE_INIT:
    // disable and re-enable fifo, see datasheet
    rf12_cmd(RF12_FIFO_RESET, RF12_FIFO_IT_LEVEL(8) | RF12_FIFO_FILL(0) | RF12_DR);
    rf12_cmd(RF12_FIFO_RESET, RF12_FIFO_IT_LEVEL(8) | RF12_FIFO_FILL(0) | RF12_DR | RF12_FF);
    pkt_state.state = STATE_IDLE;
    break;

  case STATE_IDLE:
    if(rf12_read_status_MSB()) { // fifo has data
      pkt_state.len = rf12_RX()^0xAA;
      if(pkt_state.len < PKT_LEN_MIN || pkt_state.len > PKT_LEN_MAX) {
	stats.e_bad_len++;
	pkt_state.state = STATE_INIT;
	break;
      }
      pkt_state.i = 0;
      pkt_state.cksum = pkt_state.len;
      pkt_state.timeout = RCV_TIMEOUT;
      pkt_state.state = STATE_READING;
      pkt_state.buf = buffers[rbuf];
    }
    break;

  case STATE_READING:
    if(!--pkt_state.timeout) {
      stats.e_to_reading++;
      pkt_state.state = STATE_INIT;
      break;
    }
    if(rf12_read_status_MSB()) {
      pkt_state.buf[pkt_state.i] = rf12_RX();
      pkt_state.cksum += pkt_state.buf[pkt_state.i++];

      if(pkt_state.i == pkt_state.len) {
	pkt_state.state = STATE_READING_DONE;
      }
    }
    break;

  case STATE_READING_DONE:
    if(!--pkt_state.timeout) {
      stats.e_to_reading_done++;
      pkt_state.state = STATE_INIT;
      break;
    }
    if(rf12_read_status_MSB()) {
      rcv_cksum = rf12_RX();
      if(rcv_cksum != pkt_state.cksum) {
	stats.e_cksum++;
	pkt_state.i = 0;
      } else {
	stats.ok++;
	rxStatusLED = STATUS_LED_COUNTER;
	buffers_len[rbuf] = pkt_state.len;
	rbuf++; rbuf &= BUFFERMASK;
	if(ubuf == rbuf) { // overrun
	  ubuf++; ubuf &= BUFFERMASK;
	}
	usbSetInterrupt(0,0); // signal the host that data is ready 
      }
      pkt_state.state = STATE_INIT;
    }
    break;
  }
}
      


uchar usbFunctionSetup(uchar data[8]) {
  uchar   len = 0;

  usbMsgPtr = replyBuffer;
  if(data[1] == 0) {		// ECHO
    len = 2;
    replyBuffer[0] = data[2];
    replyBuffer[1] = data[3];
  } else if (data[1] == 1){	// get status
    len = 2;		        // RFM12 status is read in the main loop
  } else if (data[1] == 2) {	// read statistics
    len = sizeof(stats);
    usbMsgPtr = (uint8_t *)&stats;
  } else if (data[1] == 3) {	// read data
    if(ubuf == rbuf) // queue empty
      len = 0;
    else {
      usbMsgPtr = buffers[ubuf];
      len = buffers_len[ubuf];
      ubuf++; ubuf &= BUFFERMASK;
    }
  }

  return len;
}

int main(void) {
  //  wdt_enable(WDTO_1S);
  DDRB = _BV(PB2) | _BV(PB3) | _BV(PB5); // SS, MOSI, SCK
  PORTB = _BV(PB2); // deselect RFM12

  PORTC = 0;
  DDRC = _BV(DDC2) | _BV(DDC3) | _BV(DDC4);
  PORTC = (LED_R | LED_Y | LED_G); // turn off LEDs
  RED_ON();

  PORTD = 0;
  DDRD = 0;

  usbDeviceDisconnect();
  _delay_ms(15);		/* delay >10ms for USB reset */
  usbDeviceConnect();
  RED_OFF(); YELLOW_ON();
  usbInit();
  sei();
  rf12_init();
  rf12_cmd(RF12_POWER_SETTING, RF12_ER); // enable receiver
  YELLOW_OFF(); GREEN_ON();
  for(;;){    /* main event loop */
    wdt_reset();
    usbPoll();
    if(usbStatusLED) {// flash the red LED on USB activity (cf usbconfig.h)
      GREEN_OFF();
      usbStatusLED--;
    } else
      GREEN_ON();

    if(rxStatusLED) {
      YELLOW_ON();
      rxStatusLED--;
    } else
      YELLOW_OFF();

    if(ubuf == rbuf) // queue empty
      RED_OFF();
    else
      RED_ON();

    receive_packet();
    rf12_read_status();
    replyBuffer[0] = status_H.byte;
    replyBuffer[1] = status_L.byte;

  }
  return 0;
}

