/*
    remote-sensor, a wireless sensor data collection system
    Copyright (C) 2008  christian mock <cm@tahina.priv.at>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#define F_CPU 12000000UL

#include <avr/io.h>
#include <avr/wdt.h>
#include <avr/pgmspace.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "usbdrv.h"
#include "rf12.h"

// leds: PC2 R
//       PC3 Y
//       PC4 G
#define LED_R (_BV(PC2))
#define LED_Y (_BV(PC3))
#define LED_G (_BV(PC4))
#define RED_OFF() (PORTC |= LED_R)
#define YELLOW_OFF() (PORTC |= LED_Y)
#define GREEN_OFF() (PORTC |= LED_G)
#define RED_ON() (PORTC &= ~LED_R)
#define YELLOW_ON() (PORTC &= ~LED_Y)
#define GREEN_ON() (PORTC &= ~LED_G)


static uchar replyBuffer[20];

// min and max size of received packets
#define PKT_LEN_MIN 2
#define PKT_LEN_MAX 8

// state machine for reception of wireless packets
// (we can't just do a blocking read, since usbPoll() must be called
// regularily)
// states:
// STATE_INIT: initialize the SM
// STATE_IDLE: we're waiting for the beginning of a packet
// STATE_PKT_READING: we're reading len bytes
// STATE_PKT_READING_DONE: we've read all data bytes, wait for the final cksum
#define STATE_INIT 0
#define STATE_IDLE 1
#define STATE_READING 2
#define STATE_READING_DONE 3

// one iteration of the main loop takes about 35s,
// one packet of 20bytes @9600 takes about 19ms, so...
#define RCV_TIMEOUT 600

struct {
  uint16_t freq;
  uint8_t rssi, nrssi;
} specdata;

uint16_t usbStatusLED; // counter to flash the red LED on USB activity
uint16_t rxStatusLED; // same for radio rcv

volatile uint8_t start;

uchar usbFunctionSetup(uchar data[8]) {
  uchar   len = 0;

  usbMsgPtr = replyBuffer;
  if(data[1] == 0) {		// ECHO
    len = 2;
    replyBuffer[0] = data[2];
    replyBuffer[1] = data[3];
  } else if (data[1] == 1){	// start
    len = 0;	
    start = 1;
  } 

  return len;
}

int main(void) {
  uint8_t iter;
#define NITER 20

  DDRB = _BV(PB2) | _BV(PB3) | _BV(PB5); // SS, MOSI, SCK
  PORTB = _BV(PB2); // deselect RFM12

  PORTC = 0;
  DDRC = _BV(DDC2) | _BV(DDC3) | _BV(DDC4);
  PORTC = (LED_R | LED_Y | LED_G); // turn off LEDs
  RED_ON();

  PORTD = 0;
  DDRD = 0;

  usbDeviceDisconnect();
  _delay_ms(15);		/* delay >10ms for USB reset */
  usbDeviceConnect();
  RED_OFF(); YELLOW_ON();
  usbInit();
  sei();
  rf12_init();
  rf12_cmd(RF12_POWER_SETTING, RF12_ER); // enable receiver
  rf12_cmd(RF12_AFC, 0); // disable AFC
  YELLOW_OFF(); GREEN_ON();
  for(;;){    /* main event loop */
    YELLOW_ON(); RED_ON();
    while(!start) // wait for start signal over USB
      usbPoll();
    YELLOW_OFF(); RED_OFF();
    for(specdata.freq = 96; specdata.freq <= 3093; specdata.freq+=20) {
      // 100kHz steps
      rf12_cmd(RF12_FREQ_SETTING_HIGH(specdata.freq),
	       RF12_FREQ_SETTING_LOW(specdata.freq));
      for(specdata.rssi = 0; specdata.rssi <= 7; specdata.rssi++) {
	// -103 .. -61 dBm
	rf12_cmd(RF12_RECEIVER_CONTROL(RF12_VDI_FAST | RF12_P20_VDI),
		 RF12_BW_400 | RF12_LNA_0dB | specdata.rssi);
	specdata.nrssi = 0;
	YELLOW_ON();
	for(iter = NITER; iter; iter--) {
	  _delay_ms(1);
	  usbPoll();
	  rf12_read_status();
	  if(status_H.bits.ATS_RSSI)
	    specdata.nrssi++;
	}
	YELLOW_OFF(); RED_ON();
	while(!usbInterruptIsReady()) // wait for host
	  usbPoll();
	usbSetInterrupt(&specdata,sizeof(specdata));
	RED_OFF();
      }
    }
    start = 0;
  }
  return 0;
}

