/*
    remote-sensor, a wireless sensor data collection system
    Copyright (C) 2008  christian mock <cm@tahina.priv.at>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*/

#include <stdio.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <sys/errno.h>
#include <math.h>
#include <usb.h> /* written for libusb 0.1.12 */
#include "sensord.h"

struct {
  uint16_t freq;
  uint8_t rssi, nrssi;
} specdata; // __attribute__((packed));

int rssitab[] = { -103, -97, -91, -85, -79, -73, -67, -61 };

struct usb_bus *busses;

usb_dev_handle *find_sensor(void) {
  struct usb_bus *bus;
  char Manufacturer[STRINGSZ], Product[STRINGSZ];
  struct usb_device *dev;
  struct usb_dev_handle *devh = NULL;

  usb_init();
  usb_find_busses();
  usb_find_devices();
  busses = usb_get_busses();

  for(bus = busses; bus; bus = bus->next) {
    for(dev = bus->devices; dev; dev = dev->next) {
      if(dev->descriptor.idVendor == USB_VENDOR &&
	 dev->descriptor.idProduct == USB_PRODUCT &&
	 dev->descriptor.bcdDevice == USB_PRODUCT_VERSION) {
	devh = usb_open(dev);
	if(!devh) {
	  perror("usb_open");
	  goto xit;
	}
	if(usb_get_string_simple(devh, dev->descriptor.iManufacturer,
				 Manufacturer, STRINGSZ-1) < 0) {
	  perror("usb_get_string_simple(iManufacturer)");
	  goto xit;
	}
	if(usb_get_string_simple(devh, dev->descriptor.iProduct, Product,
				 STRINGSZ-1) < 0) {
	  perror("usb_get_string_simple(iProduct)");
	  goto xit;
	}
	if(!strcmp(Manufacturer, USB_MANUF_STRING) &&
	   !strcmp(Product, USB_PRODUCT_STRING)) {
	  if(usb_set_configuration(devh, 1) < 0) {
	    perror("usb_set_configuration"); exit(1);
	  }
	  if(usb_claim_interface(devh, 0) < 0) {
	    perror("usb_claim_interface"); exit(1);
	  }
	  if((dev->config->interface->altsetting->endpoint->bmAttributes &
	      USB_ENDPOINT_TYPE_MASK) != USB_ENDPOINT_TYPE_INTERRUPT) {
	    fprintf(stderr, "Endpoint 1 is not an interrupt endpoint\n");
	    goto xit;
	  }
	  return devh;
	}
      }
    xit:
      if(devh) {
	usb_release_interface(devh, 0);
	usb_close(devh);
	devh = NULL;
      }
    }
  }
  return NULL;
}

void close_sensor(usb_dev_handle *devh) {
  if(devh) {
    usb_release_interface(devh, 0);
    usb_close(devh);
  }
}

int main(int argc, char **argv) {
  usb_dev_handle *devh;
  int len;
  int l_freq, max_rssi = -1000;

  if(!(devh = find_sensor())) {
    fprintf(stderr, "No receiver found\n");
    exit(1);
  }

  len = usb_control_msg(devh, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, 1, 0, 0, NULL, 0, 1000);


  while(1) {
    if((len = usb_interrupt_read(devh, 1, &specdata, sizeof(specdata), 1000)) < 0) {
      //      if(errno != EAGAIN) {
	fprintf(stderr, "usb_interrupt_read: %s\n", usb_strerror());
	close_sensor(devh);
	exit(2);
	//      }
    } else {
      if(specdata.freq != l_freq) {
	if(max_rssi != -1000)
	  printf("%.3f %d\n", 20.0 * (43 + l_freq/4000.0),
		 max_rssi);
	max_rssi = -1000;
      }
      if(specdata.nrssi)
	max_rssi = rssitab[specdata.rssi] - 6 + specdata.nrssi/20*6;
      l_freq = specdata.freq;
    }
  }

  close_sensor(devh);
}
    
